/* 
 * secsplit.c
 *
 * Shamir secret sharing
 *
 * Based on "How to Share a Secret", by Adi Shamir, Communications of
 * the ACM, November, 1979, Volume 22, Number 11, page 612.
 *
 * Original copyright:
 * Copyright (C) 1993 Hal Finney, 74076.1041@compuserve.com.
 * This software is being placed in the public domain.
 *
 * This program divides a file into n pieces, such that any k of them
 * are sufficient to reconstruct the original file, but that k-1
 * pieces give NO information about the original file (except its
 * length).
 *
 * It has been written for and tested on DOS and Unix systems.
 *
 * To split up a file, use the command; secsplit n k file where n is
 * the number of pieces to split it up to, and k is the number of
 * pieces needed to reconstruct it. k should be <= n and > 1.  If you
 * get n and k backwards the program will swap them for you so don't
 * worry too much about remembering the order.
 *
 * The program will output to file.001, file.002, .... If the file has
 * an extension (e.g. "file.c") the extension will be stripped off
 * before the ".001", etc., are added (so "file.c" will also output to
 * "file.001", etc.).
 *
 * To reconstruct a file, use the command:
 *    secsplit k file.*
 * or
 *    secsplit k file file1 file2...
 *
 * k should be >= the k used when the file was split; the minimum
 * number of pieces needed to reconstruct the file.  If you have too
 * few pieces then the program won't give an error, but you'll get the
 * wrong answer.
 *
 * The first command form is for DOS or other systems which won't
 * expand the ".*" for you; the program scans for file.000, file.001,
 * etc., and uses the first k of them that it finds.  In the second
 * form, the number of files given should be at least k, and again the
 * first k of them will be used.
 *
 * The output in the first form will be file.out; in the second form
 * it will be the first file on the command line, stripped of its
 * extension, and with ".out" added, so generally it will be file.out
 * too.
 *
 * Shamir's algorithm relies on cryptographically strong, unguessable,
 * random numbers.  This version of the program uses the /dev/urandom
 * interface to the kernel random number pool present on the free
 * Unices.
 *
 *
 * Revision history:
 *    Version 1.0	October 23, 1993
 *
 *    Version 1.1	October 24, 1993
 *       Added IDEA-based random-number-generator, initialized by MD5
 *       of input file, plus the time of day.
 *
 *    Version 1.2	March 31, 2000 Damien Miller <djm@mindrot.org>
 *       Removed IDEA-based random-number generator in favour of 
 *       /dev/random. Removed MSDOSisms. Rearranged source code and 
 *       formatting according to OpenBSD style(9)
 *
 *
 * The file formats used for output are as follows.  Each file starts
 * with one byte which is the index, from 1 through n, of that file.
 * This is the x value used for the polynomial evaluation in Shamir's
 * algorithm.  The files then consist of a series of 16-bit values
 * (high byte first), which are the result of applying Shamir's
 * splitting algorithm to the input file taken in 16-bit chunks.  The
 * prime used is slightly less than 2^16, meaning that input data
 * values close to 2^16 get turned into a pair of values (see below
 * for more information on this expansion).
 *
 * If the input file is of even length, the output files will each by
 * of that length+1 (because of the 1 byte at the beginning).  If the
 * input file is of odd length, the input is padded with a random byte
 * and processed normally to get a pair of output bytes, then each
 * output file is padded with an extra random byte to indicate this
 * fact.  So output files which have an odd length correspond to input
 * files with an even length, and vice versa.
 *
 * The output files could have encoded k and n information, but this
 * could be helpful to an attacker (he would know when he was close to
 * having enough pieces to reconstruct the file if he knew k).  So the
 * user is required to remember this himself.
 *
 * The file formats are system-independent so files split on one kind
 * of machine should be able to be reassembled on another kind of
 * machine.
 *
 * One complication exists with this code.  The algorithm must work
 * mod a prime which is bigger than all the data.  We want the data to
 * be about 16 bits therefore, so that we can multiply two numbers and
 * have them fit in a 32-bit int.  This means that the prime has to be
 * less than 2^16 in order for them to fit.  The largest such prime is
 * 65521.  This means that any data value > 65521 can't be processed
 * directly; instead, I split it into two values, a 65520, followed by
 * value-65520.  Also, 65520 itself is split as 65520 followed by 0.
 * This complication is pretty well limited to the input routine on
 * splitting and the output routine on assembly.
 *
 * The purpose of the "magic" is to xor the incoming value with a
 * different number each time.  This is purely intended to deal with
 * the problem that values of d >= limit-1 must be stored as four
 * bytes rather than two.  I was worried that some file might have a
 * lot of 0xff's which would balloon in size.  So I wanted to xor with
 * a rather random number.  Then, I change the random number each time
 * in order to make it less likely that some other file would happen
 * to have a lot of ballooning values.  This way it would be very rare
 * that a file managed to track my magic number in such a way that
 * many d's were over the size threshold.
 *
 * (The "magic" is not intended to increase the security or the
 * cryptographic strength of the algorithm in any way; it is purely to
 * keep the size of the output files from being much bigger than the
 * input.)
 *
 */

#include <sys/types.h>
#include <sys/stat.h>

#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#define PRIME 65521
#define KMAX 48
#define NMAX 48

#ifdef NOMAGIC
/* For debugging */
#define IMAGIC 0
#define DMAGIC 0
#else
/* Two random values - see comment above to explain magic */
#define IMAGIC 0x8a31
#define DMAGIC 0x1347
#endif

#define RANDOM_POOL "/dev/urandom"

/* Multiplicative inverses of 1-48 mod PRIME */
int invtab[] = {
	    1, 32761, 43681, 49141, 52417, 54601, 56161, 57331,
	58241, 58969, 11913, 60061, 60481, 60841, 61153, 61426,
	42396, 61881,  6897, 62245, 62401, 38717, 11395, 62791,
	49796, 63001, 41254, 63181, 58743, 63337, 25363, 30713,
	 3971, 21198, 63649, 63701, 54896, 36209, 63841, 63883,
	43148, 63961,  6095, 52119, 64065, 38458, 43216, 64156
};

/******************* Code related to splitting *********************/

/*
 * Return a random number from 0 to n-1. 
 */
unsigned int
crandom(unsigned int n)
{
	static int random_fd = -1;
	unsigned int c;

	if (random_fd == -1)
		random_fd = open(RANDOM_POOL, O_RDONLY);

	if (random_fd == -1) {
		fprintf(stderr, "Couldn't open random pool \"%s\": %s", RANDOM_POOL, 
			strerror(errno));
		exit(1);
	}

	if (read(random_fd, &c, sizeof(c)) != sizeof(c)) {
		fprintf(stderr, "Couldn't read from random pool \"%s\": %s", RANDOM_POOL, 
			strerror(errno));
		exit(1);
	}

	c %= n;

	return(c);
}

/*
 * Evaluate the given polynomial, n coefficients, at point x=i.
 * Do it mod the specified modulus.
 *
 * poly = Polynomial coefficients
 * n = # coefficients (order of polynomial + 1)
 * i = Point to evaluate it at
 * mod = Modulus for evaluation
 */
unsigned int
eval(unsigned int *poly, int n, int i, unsigned int mod)
{
	u_int32_t prod;		/* Accumulated product */
	int j;					/* index */

	prod = poly[n - 1];
	
	for (j = (n - 1); --j >= 0; ) {
		prod *= i;
		prod += poly[j];
		prod %= mod;
	}

	return(prod);
}

/*
 * Return a 16-bit value from file f_in, but limit it to be less than limit.
 * Anything >= limit-1 gets returned as two consecutive values (on 2 calls).
 * Return -1 on EOF, or -2 if the previous return value had been padded
 * because the file had an odd # bytes.
 */
unsigned int
get_limited_16 (FILE *f_in, unsigned int limit)
{
	static int have_extra;
	static unsigned int extra;
	static int oddflag;
	static unsigned int magic = IMAGIC;
	unsigned int c1, c2;
	unsigned int d;

	/* First check for leftover from last time */
	if (have_extra) {
		have_extra = 0;
		return(extra);
	}

	/* Check if last return included a pad */
	if (oddflag)
		return(-2);

	/* Read data (bigendian), do the magic */
	c1 = fgetc(f_in);
	if (c1 == EOF)
		return(-1);

	c2 = fgetc(f_in);
	if (c2 == EOF) {
		c2 = crandom(0x100);
		oddflag = 1;
	}

	d = ((c1 & 0xff) << 8) + (c2 & 0xff);
	d ^= magic;
	magic = (magic + DMAGIC) & 0xffff;

	/*
	 * If over the limit, return limit-1 as a code for that, and remember
	 * to return the rest next time.
	 */
	if (d >= (limit - 1)) {
		have_extra = 1;
		extra = d - (limit - 1);
		d = limit - 1;
	}
	return(d);
}

/*
 * Given a 16-bit value d, less than mod, split it into nout files such
 * that any k of them can reconstruct it.
 */
void
split_out (unsigned int d, FILE *f_out[], int nout, int k, unsigned int mod)
{
	unsigned int poly[KMAX];
	int i, j;
	unsigned int di;

	poly[0] = d;
	for (j = 1; j < k; ++j) {
		poly[j] = crandom(mod);
	}

	for (i=0; i < nout; ++i) {
		di = eval(poly, k, i + 1, mod);
		fputc((di >> 8) & 0xff, f_out[i]);
		fputc(di & 0xff, f_out[i]);
	}
}

/*
 * Split the specified input file into nout output files, such that
 * any k of them are sufficient to reconstruct the input.  mod is the
 * largest prime < 2^16.  This is the main routine for the splitting case.
 *
 * f_in = Input file handle
 * f_out[] = Output file handles
 * nout = Number of output files
 * k = Threshhold for re-assembl
 * mod = Modulus for calculations
 */
void
split(FILE *f_in, FILE *f_out[], int nout, int k, unsigned int mod)
{
	int i;
	unsigned int d;

	/* Prefix each file with "x" coordinate, 1 byte */
	for (i = 0; i < nout; ++i)
		fputc(i + 1, f_out[i]);
	
	for ( ; ; ) {
		d = get_limited_16(f_in, PRIME);
		if (d == -1)
			break;

		if (d == -2) {
			/* Odd flag - pad output files with a random byte to remember */
			for (i = 0; i < nout; ++i)
				fputc(crandom(0x100), f_out[i]);

			break;
		}
		split_out(d, f_out, nout, k, mod);
	}
}

/******************* Code related to assembly *********************/

/*
 * Return the multiplicative inverse of small positive or negative value
 * x modulo the current prime.  Do a table lookup for speed. 
 *
 * x = Small value to find inverse of
 * mod = Argument is ignored
 */
unsigned int
inverse(int x, unsigned int mod)
{
	int neg = 0;

	if (x < 0) {
		x = -x;
		neg = 1;
	}

	if ((x < 1) || (x > (sizeof(invtab)/sizeof(invtab[0])))) {
		fprintf (stderr, "inverse out of range: %d\n", x);
		exit(1);
	}

	if (neg)
		return(PRIME - invtab[x-1]);
	else
		return(invtab[x-1]);
}


/*
 ( Interpolate the polynomial specified at x and y coordinates
 * in the array of size n, at x=i.  Do it mod the specified modulus.
 * This algorithm is from Knuth, The Art of Computer Programming, Vol. 2,
 * Seminumerical Algorithms, section 4.6.4, Evaluation of Polynomials,
 * equation 43 and 44, page 485 (bottom) in 1981 hardcover edition.
 * Knuth has his indices go from 0 to n; mine go from 0 to n-1, a slight
 * notational change.
 *
 * i = x coord of interpolated point
 * x[] = x coordinates of known points
 * y[] = y coordinates of known points
 * n = size of x, y, and alpha arrays
 * mod = modulus for reducing results
 */
unsigned int 
interp(int i, int x[], unsigned int y[], int n, unsigned int mod)
{
	u_int32_t alpha[KMAX];
	int j, k;
	u_int32_t prod;

	for (j = 0; j < n; ++j) {
		alpha[j] = y[j];
#ifdef DEBUG
		printf ("Interp: alpha[%d] = %x\n", j, alpha[j]);
#endif
	}

	for (k = 1; k < n; ++k) {
		for (j = n - 1; j >= k; --j) {
			if (alpha[j] > alpha[j - 1])
				alpha[j] = alpha[j] - alpha[j - 1];
			else
				alpha[j] = alpha[j] - alpha[j - 1] + mod;

			alpha[j] *= inverse (x[j] - x[j - k], mod);
			alpha[j] = alpha[j] % mod;
		}
	}
#ifdef DEBUG
	for (j = 0; j < n; ++j)
		printf("Interp: alpha[%d] = %x\n", j, alpha[j]);
#endif

	prod = alpha[n - 1];

	for(j = n - 2; j >= 0; --j) {
		if (i < x[j]) {
			prod *= i - x[j] + mod;
		} else {
			prod *= i - x[j];
		}
		prod += alpha[j];
		prod %= mod;
	}
#ifdef DEBUG
	printf ("Interp: prod = %x\n", prod);
#endif
	return(prod);
}


/*
 * Return a 16-bit-value from the file, with a flag set if only the
 * high 8 bits of the reconstructed value will be valid.  This is known
 * because odd-size files are padded with an extra byte.  We have to stay
 * two bytes ahead to know this.
 */
unsigned int 
get_assemble_16(FILE *f_in[], int i, int *podd)
{
	static int notfirst[KMAX];
	static unsigned int next_d1[KMAX];
	unsigned int d1;
	unsigned int c1, c2;

	if (!notfirst[i]) {
		/* Get ahead the first time */
		c1 = fgetc(f_in[i]);
		c2 = fgetc(f_in[i]);
		next_d1[i] = ((c1 & 0xff) << 8) + (c2 & 0xff);
		notfirst[i] = 1;
	}

	d1 = next_d1[i];
	c1 = fgetc(f_in[i]);

	if (c1 == EOF) {
		next_d1[i] = -1;
	} else {
		c2 = fgetc(f_in[i]);
		if (c2 == EOF) {
			*podd = 1;
			next_d1[i] = -1;
		} else {
			next_d1[i] = ((c1 & 0xff) << 8) + (c2 & 0xff);
		}
	}
	return(d1);
}
    
    
/*
 * Get 16-bit values from each file, and interpolate them to get the value
 * at x=0.  That is how we do the assembly.  Set the podd flag if
 * only the high 8 bits of this interpolated value are valid.
 */
unsigned int
get_assemble(FILE *f_in[], int nin, int x[], FILE *f_out, unsigned int mod, 
	int *podd)
{
	unsigned int y[KMAX];
	int i;

	for (i = 0; i < nin; ++i) {
		y[i] = get_assemble_16(f_in, i, podd);
		if (y[i] == -1) /* EOF */
			return -1;
	}
	return(interp(0, x, y, nin, mod));
}

/*
 * Given a set of nin file descriptors, assemble them to generate the
 * original file.  This is the main routine for the assembly case.
 */
void
assemble(FILE *f_in[], char *names[], int nin, FILE *f_out, unsigned int mod)
{
	unsigned int magic = IMAGIC;
	int i;
	int x[KMAX];
	int oddflag = 0;
	unsigned int c1, c2;
	unsigned int d;

	/* Read x index number from each file, one byte per */
	for(i = 0; i < nin; ++i) {
		x[i] = fgetc(f_in[i]);
		if (x[i] > NMAX) {
			fprintf (stderr, "File %s does not appear to be valid\n", names[i]);
			exit (1);
		}
	}

	for ( ; ; ) {
		d = get_assemble (f_in, nin, x, f_out, mod, &oddflag);
		if (d == -1)
			break;

		if (d == mod-1) {
			d = get_assemble(f_in, nin, x, f_out, mod, &oddflag);
			d += mod - 1;
		}

		d ^= magic;
		magic = (magic + DMAGIC) & 0xffff;
		c1 = (d >> 8) & 0xff;
		c2 = d & 0xff;
		fputc (c1, f_out);

		if (oddflag)
			break;

		fputc (c2, f_out);
	}
}

/********************************* MAIN *********************************/

char *av0;

void 
usage(void)
{
	fprintf(stderr, "Usage: %s <n-pieces> <n-needed> file\n", av0);
	fprintf(stderr, "   (splits to file.001, file.002, ...)\n");
	fprintf(stderr, "or %s <n-files> file.*\n", av0);
	fprintf(stderr, "or %s <n-files> file file2 [...]\n", av0);
	fprintf(stderr, "   (These last two re-assemble the pieces to file.out)\n");
	exit(1);
}

void 
*xmalloc(size_t n)
{
	void *result;
	
	result = malloc(n);
	if (result == NULL) {
		fprintf(stderr, "Out of memory.\n");
		exit(1);
	}
	
	return(result);
}

void 
split_from(char *secret, int num_shadows, int shadows_required)
{
	static FILE *f_in;
	static FILE *f_out[NMAX];
	int fnamelen;
	static char fname[1024];
	int i;
	char *fnp;
	static char *names[NMAX];

	f_in = fopen(secret, "r");
	if (f_in == NULL) {
		fprintf (stderr, "Unable to open file %s\n", secret);
		exit(1);
	}

	/* Strip off extension */
	snprintf(fname, sizeof(fname), "%s", secret);
	fnamelen = strlen(fname);

	for (fnp = fname + fnamelen; (*--fnp != '.') && (fnp > fname); )
		;	/* Nothing */

	if (fnp != fname) {
		*fnp = '\0';
		fnamelen = fnp - fname;
	}

	/* Open output files */
	for (i = 0; i < num_shadows; i++) {
		names[i] = (char*)xmalloc(fnamelen + 5);
		sprintf(names[i], "%s.%03d", fname, i + 1);
		
		f_out[i] = fopen(names[i], "w");
		if (f_out[i] == NULL) {
			fprintf(stderr, "Unable to create file %s\n", names[i]);
			exit(1);
		}
	}

	printf ("Splitting file %s to:\n", secret);
	for (i = 0; i < num_shadows; ++i) {
		printf("%s%s", names[i], (i == (num_shadows - 1))?"\n":" ");
	}

	/* Do the work */
	split(f_in, f_out, num_shadows, shadows_required, PRIME);

	/* Clean up and exit */
	for (i = 0; i < num_shadows; ++i)
		fclose(f_out[i]);

	fclose(f_in);

	exit(0);
}

void 
join_from(char **files, int num_files, int num_shadows)
{
	static FILE *f_in[NMAX];
	static FILE *f_out;
	int fnamelen;
	static char fname[1024];
	char *fnp;
	static char *names[NMAX];
	int i;
	int n;

	n = num_shadows;

	/* filename list */
	if (num_files < num_shadows) {
		fprintf(stderr, "Insufficient files specified\n");
		exit(1);
	}

	/* Open input files */
	for (i = 0; i < num_shadows; ++i) {
		names[i] = files[i];

		f_in[i] = fopen(names[i], "r");
		if (f_in[i] == NULL) {
			fprintf(stderr, "Unable to open file %s\n", names[i]);
			exit(1);
		}
	}

	/* Strip extension off 1st input file */
	snprintf(fname, sizeof(fname) - 4, "%s", files[0]);
	fnamelen = strlen(fname);

	for (fnp = fname + fnamelen; (*--fnp!='.') && (fnp > fname); )
	;	/* Nothing */

	if (fnp != fname) {
		*fnp = '\0';
		fnamelen = fnp - fname;
	}

	/* Two assembly cases (fname-list and glob) come together here */
	/* Add .out and open output files */
	strcat(fname, ".out");
	f_out = fopen(fname, "w");
	if (f_out == NULL) {
		 fprintf(stderr, "Unable to create output file %s\n", fname);
		 exit(1);
	}

	printf("Creating file %s from files:\n", fname);
	for (i = 0; i < num_shadows; ++i)
		printf("%s%s", names[i], (i == (num_shadows - 1))?"\n":" ");

	/* Do the work */
	assemble(f_in, names, num_shadows, f_out, PRIME);

	/* Clean up and exit */
	for (i = 0; i < n; i++)
		fclose(f_in[i]);

	fclose(f_out);

	exit(0);
}
            
int 
main(int argc, char **argv)
{
	int num_shadows;
	int shadows_required;

	av0 = argv[0];
	if (argc < 3)
		usage();

	num_shadows = atoi(argv[1]);
	if ((num_shadows <= 1) || (num_shadows > NMAX)) {
		fprintf(stderr, "Invalid number of shadows.\n\n");
		usage();
	}
	
	shadows_required = atoi(argv[2]);

	if (shadows_required == 0)
		join_from(argv + 2, argc - 2, num_shadows);
	
	if ((shadows_required <= 1) || (shadows_required > NMAX)) {
		fprintf(stderr, "Invalid number of shadows required.\n\n");
		usage();
	}

	if (argc != 4) {
		fprintf(stderr, "Invalid number of commandline arguments.\n\n");
		usage();
	}

	if (shadows_required > num_shadows) {
		fprintf(stderr, "Number of requires pieces is greater than number of pieces.\n\n");
		usage();
	}
	
	split_from(argv[3], num_shadows, shadows_required);
	
	/* NOTREACHED */
	exit(0);
}
